@@ -33,14 +33,18 @@ struct connman_network {
 	struct connman_element element;
 	enum connman_network_type type;
 	enum connman_network_protocol protocol;
+	connman_bool_t associating;
+	connman_bool_t secondary;
 	connman_bool_t available;
 	connman_bool_t connected;
 	connman_bool_t remember;
 	connman_uint8_t priority;
 	connman_uint8_t strength;
 	char *identifier;
+	char *address;
 	char *name;
 	char *node;
+	char *group;
 
 	struct connman_network_driver *driver;
 	void *driver_data;
@@ -109,6 +113,10 @@ static DBusMessage *get_properties(DBusConnection *conn,
 						DBUS_TYPE_OBJECT_PATH, &path);
 	}
 
+	if (network->address != NULL)
+		connman_dbus_dict_append_variant(&dict, "Address",
+					DBUS_TYPE_STRING, &network->address);
+
 	if (network->name != NULL)
 		connman_dbus_dict_append_variant(&dict, "Name",
 					DBUS_TYPE_STRING, &network->name);
@@ -232,7 +240,7 @@ static DBusMessage *do_connect(DBusConnection *conn,
 		return __connman_error_permission_denied(msg);
 
 	if (network->connected == TRUE)
-		return __connman_error_failed(msg);
+		return __connman_error_failed(msg, EALREADY);
 
 	if (network->driver && network->driver->connect) {
 		enum connman_device_mode mode;
@@ -243,7 +251,7 @@ static DBusMessage *do_connect(DBusConnection *conn,
 
 		err = network->driver->connect(network);
 		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg);
+			return __connman_error_failed(msg, -err);
 	} else
 		network->connected = TRUE;
 
@@ -263,7 +271,7 @@ static DBusMessage *do_disconnect(DBusConnection *conn,
 		return __connman_error_permission_denied(msg);
 
 	if (network->connected == FALSE)
-		return __connman_error_failed(msg);
+		return __connman_error_failed(msg, EINVAL);
 
 	connman_element_unregister_children(&network->element);
 
@@ -272,7 +280,7 @@ static DBusMessage *do_disconnect(DBusConnection *conn,
 	if (network->driver && network->driver->disconnect) {
 		err = network->driver->disconnect(network);
 		if (err < 0 && err != -EINPROGRESS)
-			return __connman_error_failed(msg);
+			return __connman_error_failed(msg, -err);
 	} else
 		network->connected = FALSE;
 
@@ -428,8 +436,10 @@ static void network_destruct(struct connman_element *element)
 	g_free(network->wifi.security);
 	g_free(network->wifi.passphrase);
 
+	g_free(network->group);
 	g_free(network->node);
 	g_free(network->name);
+	g_free(network->address);
 	g_free(network->identifier);
 }
 
@@ -448,7 +458,6 @@ struct connman_network *connman_network_create(const char *identifier,
 	connman_uint8_t strength = 0;
 	const char *str;
 	char *temp;
-	unsigned int i;
 
 	DBG("identifier %s type %d", identifier, type);
 
@@ -460,19 +469,13 @@ struct connman_network *connman_network_create(const char *identifier,
 
 	__connman_element_initialize(&network->element);
 
+	//temp = connman_dbus_encode_string(identifier);
 	temp = g_strdup(identifier);
 	if (temp == NULL) {
 		g_free(network);
 		return NULL;
 	}
 
-	for (i = 0; i < strlen(temp); i++) {
-		char tmp = temp[i];
-		if ((tmp < '0' || tmp > '9') && (tmp < 'A' || tmp > 'Z') &&
-						(tmp < 'a' || tmp > 'z'))
-			temp[i] = '_';
-	}
-
 	network->element.name = temp;
 	network->element.type = CONNMAN_ELEMENT_TYPE_NETWORK;
 
@@ -481,13 +484,12 @@ struct connman_network *connman_network_create(const char *identifier,
 
 	str = type2string(type);
 	if (str != NULL)
-		connman_element_set_static_property(&network->element,
-					"Type", DBUS_TYPE_STRING, &str);
+		connman_element_set_string(&network->element, "Type", str);
 
-	connman_element_set_static_property(&network->element,
-					"Strength", DBUS_TYPE_BYTE, &strength);
+	connman_element_set_uint8(&network->element, "Strength", strength);
 
-	network->type = type;
+	network->type       = type;
+	network->secondary  = FALSE;
 	network->identifier = g_strdup(identifier);
 
 	return network;
@@ -518,6 +520,22 @@ void connman_network_unref(struct connman_network *network)
 	connman_element_unref(&network->element);
 }
 
+const char *__connman_network_get_type(struct connman_network *network)
+{
+	return type2string(network->type);
+}
+
+/**
+ * connman_network_get_type:
+ * @network: network structure
+ *
+ * Get type of network
+ */
+enum connman_network_type connman_network_get_type(struct connman_network *network)
+{
+	return network->type;
+}
+
 /**
  * connman_network_get_identifier:
  * @network: network structure
@@ -576,6 +594,33 @@ void connman_network_set_protocol(struct connman_network *network,
 	network->protocol = protocol;
 }
 
+/**
+ * connman_network_set_group:
+ * @network: network structure
+ * @group: group name
+ *
+ * Set group name for automatic clustering
+ */
+void connman_network_set_group(struct connman_network *network,
+							const char *group)
+{
+	g_free(network->group);
+	network->group = g_strdup(group);
+}
+
+const char *__connman_network_get_group(struct connman_network *network)
+{
+	return network->group;
+}
+
+const char *__connman_network_get_ident(struct connman_network *network)
+{
+	if (network->device == NULL)
+		return NULL;
+
+	return __connman_device_get_ident(network->device);
+}
+
 /**
  * connman_network_set_available:
  * @network: network structure
@@ -639,6 +684,9 @@ connman_bool_t connman_network_get_available(struct connman_network *network)
 static gboolean set_connected(gpointer user_data)
 {
 	struct connman_network *network = user_data;
+	struct connman_service *service;
+
+	service = __connman_service_lookup_from_network(network);
 
 	if (network->connected == TRUE) {
 		struct connman_element *element;
@@ -669,8 +717,14 @@ static gboolean set_connected(gpointer user_data)
 			if (connman_element_register(element,
 						&network->element) < 0)
 				connman_element_unref(element);
+
+			__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_CONFIGURATION);
 		}
 	} else {
+		__connman_service_indicate_state(service,
+						CONNMAN_SERVICE_STATE_IDLE);
+
 		connman_element_unregister_children(&network->element);
 
 		__connman_device_set_network(network->device, NULL);
@@ -681,6 +735,34 @@ static gboolean set_connected(gpointer user_data)
 	return FALSE;
 }
 
+/**
+ * connman_network_set_associating:
+ * @network: network structure
+ * @associating: associating state
+ *
+ * Change associating state of network
+ */
+int connman_network_set_associating(struct connman_network *network,
+						connman_bool_t associating)
+{
+	DBG("network %p associating %d", network, associating);
+
+	if (network->associating == associating)
+		return -EALREADY;
+
+	network->associating = associating;
+
+	if (associating == TRUE) {
+		struct connman_service *service;
+
+		service = __connman_service_lookup_from_network(network);
+		__connman_service_indicate_state(service,
+					CONNMAN_SERVICE_STATE_ASSOCIATION);
+	}
+
+	return 0;
+}
+
 /**
  * connman_network_set_connected:
  * @network: network structure
@@ -702,6 +784,9 @@ int connman_network_set_connected(struct connman_network *network,
 
 	network->connected = connected;
 
+	if (connected == TRUE)
+		network->associating = FALSE;
+
 	if (network->registered == FALSE) {
 		g_idle_add(set_connected, network);
 		return 0;
@@ -817,6 +902,8 @@ int __connman_network_disconnect(struct connman_network *network)
 	if (network->connected == FALSE)
 		return -ENOTCONN;
 
+	connman_element_unregister_children(&network->element);
+
 	if (network->driver && network->driver->disconnect)
 		return network->driver->disconnect(network);
 
@@ -838,7 +925,10 @@ int connman_network_set_string(struct connman_network *network,
 {
 	DBG("network %p key %s value %s", network, key, value);
 
-	if (g_str_equal(key, "Name") == TRUE) {
+	if (g_str_equal(key, "Address") == TRUE) {
+		g_free(network->address);
+		network->address = g_strdup(value);
+	} else if (g_str_equal(key, "Name") == TRUE) {
 		g_free(network->name);
 		network->name = g_strdup(value);
 	} else if (g_str_equal(key, "Node") == TRUE) {
@@ -855,7 +945,7 @@ int connman_network_set_string(struct connman_network *network,
 		network->wifi.passphrase = g_strdup(value);
 	}
 
-	return 0;
+	return connman_element_set_string(&network->element, key, value);
 }
 
 /**
@@ -870,7 +960,9 @@ const char *connman_network_get_string(struct connman_network *network,
 {
 	DBG("network %p key %s", network, key);
 
-	if (g_str_equal(key, "Name") == TRUE)
+	if (g_str_equal(key, "Address") == TRUE)
+		return network->address;
+	else if (g_str_equal(key, "Name") == TRUE)
 		return network->name;
 	else if (g_str_equal(key, "Node") == TRUE)
 		return network->node;
@@ -881,7 +973,7 @@ const char *connman_network_get_string(struct connman_network *network,
 	else if (g_str_equal(key, "WiFi.Passphrase") == TRUE)
 		return network->wifi.passphrase;
 
-	return NULL;
+	return connman_element_get_string(&network->element, key);
 }
 
 /**
@@ -899,13 +991,10 @@ int connman_network_set_uint8(struct connman_network *network,
 
 	if (g_str_equal(key, "Priority") == TRUE)
 		network->priority = value;
-	else if (g_str_equal(key, "Strength") == TRUE) {
+	else if (g_str_equal(key, "Strength") == TRUE)
 		network->strength = value;
-		connman_element_set_static_property(&network->element,
-					"Strength", DBUS_TYPE_BYTE, &value);
-	}
 
-	return 0;
+	return connman_element_set_uint8(&network->element, key, value);
 }
 
 /**
@@ -925,7 +1014,7 @@ connman_uint8_t connman_network_get_uint8(struct connman_network *network,
 	else if (g_str_equal(key, "Strength") == TRUE)
 		return network->strength;
 
-	return 0;
+	return connman_element_get_uint8(&network->element, key);
 }
 
 /**
@@ -952,7 +1041,7 @@ int connman_network_set_blob(struct connman_network *network,
 			network->wifi.ssid_len = 0;
 	}
 
-	return 0;
+	return connman_element_set_blob(&network->element, key, data, size);
 }
 
 /**
@@ -974,7 +1063,7 @@ const void *connman_network_get_blob(struct connman_network *network,
 		return network->wifi.ssid;
 	}
 
-	return NULL;
+	return connman_element_get_blob(&network->element, key, size);
 }
 
 void __connman_network_set_device(struct connman_network *network,
@@ -1062,6 +1151,22 @@ static int network_probe(struct connman_element *element)
 		return err;
 	}
 
+	network->secondary = connman_device_get_secondary(network->device);
+
+	switch (network->type) {
+	case CONNMAN_NETWORK_TYPE_UNKNOWN:
+	case CONNMAN_NETWORK_TYPE_VENDOR:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+	case CONNMAN_NETWORK_TYPE_HSO:
+		break;
+	case CONNMAN_NETWORK_TYPE_WIFI:
+	case CONNMAN_NETWORK_TYPE_WIMAX:
+		if (network->secondary == FALSE)
+			__connman_profile_add_network(network);
+		break;
+	}
+
 	return 0;
 }
 
@@ -1077,18 +1182,60 @@ static void network_remove(struct connman_element *element)
 	if (network->driver == NULL)
 		return;
 
+	switch (network->type) {
+	case CONNMAN_NETWORK_TYPE_UNKNOWN:
+	case CONNMAN_NETWORK_TYPE_VENDOR:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_PAN:
+	case CONNMAN_NETWORK_TYPE_BLUETOOTH_DUN:
+	case CONNMAN_NETWORK_TYPE_HSO:
+		break;
+	case CONNMAN_NETWORK_TYPE_WIFI:
+	case CONNMAN_NETWORK_TYPE_WIMAX:
+		if (network->secondary == FALSE)
+			__connman_profile_remove_network(network);
+		break;
+	}
+
 	unregister_interface(element);
 
 	if (network->driver->remove)
 		network->driver->remove(network);
 }
 
+static void network_change(struct connman_element *element)
+{
+	struct connman_network *network = element->network;
+
+	DBG("element %p name %s", element, element->name);
+
+	if (element->state != CONNMAN_ELEMENT_STATE_ERROR)
+		return;
+
+	if (element->error != CONNMAN_ELEMENT_ERROR_DHCP_FAILED)
+		return;
+
+	if (network->connected == FALSE)
+		return;
+
+	connman_element_unregister_children(element);
+
+	connman_device_set_disconnected(network->device, TRUE);
+
+	if (network->driver && network->driver->disconnect) {
+		network->driver->disconnect(network);
+		return;
+	}
+
+	network->connected = FALSE;
+}
+
 static struct connman_driver network_driver = {
 	.name		= "network",
 	.type		= CONNMAN_ELEMENT_TYPE_NETWORK,
 	.priority	= CONNMAN_DRIVER_PRIORITY_LOW,
 	.probe		= network_probe,
 	.remove		= network_remove,
+	.change		= network_change,
 };
 
 static int network_init(struct connman_device *device)
@@ -1104,6 +1251,7 @@ static int network_load(struct connman_network *network)
 	gchar *pathname, *data = NULL;
 	gsize length;
 	const char *name;
+	char *str;
 	int val;
 
 	DBG("network %p", network);
@@ -1141,13 +1289,19 @@ static int network_load(struct connman_network *network)
 	if (val > 0)
 		network->priority = val;
 
-	g_free(network->wifi.security);
-	network->wifi.security = g_key_file_get_string(keyfile,
+	str = g_key_file_get_string(keyfile,
 				network->identifier, "WiFi.Security", NULL);
+	if (str != NULL) {
+		g_free(network->wifi.security);
+		network->wifi.security = str;
+	}
 
-	g_free(network->wifi.passphrase);
-	network->wifi.passphrase = g_key_file_get_string(keyfile,
+	str = g_key_file_get_string(keyfile,
 				network->identifier, "WiFi.Passphrase", NULL);
+	if (str != NULL) {
+		g_free(network->wifi.passphrase);
+		network->wifi.passphrase = str;
+	}
 
 	g_key_file_free(keyfile);
 
@@ -1204,7 +1358,8 @@ update:
 
 	data = g_key_file_to_data(keyfile, &length, NULL);
 
-	g_file_set_contents(pathname, data, length, NULL);
+	if (g_file_set_contents(pathname, data, length, NULL) == FALSE)
+		connman_error("Failed to store network information");
 
 done:
 	g_free(data);
